/*########################################################################

  Demo program that shows how to use the FVS library to get the ridge
  direction of a fingerprint image. At the moment the library supports
  the import of bitmap images. The result is exported as a bitmap image
  as well.

  Link this sample against the FVS library (libfvs)

########################################################################*/


#include "fvs.h"


static void ExportImage(const FvsImage_t image, const FvsString_t filename)
{
    FvsFile_t file = FileCreate();
    if (file!=NULL)
    {
        fprintf(stdout, "exporting resulting image to %s\n", filename);
        if (FileOpen(file, filename, FvsFileCreate|FvsFileWrite)==FvsOK)
        {
            (void)BitmapExport(file, image);
            (void)FileClose   (file);
        }
        FileDestroy(file);
    }
}

static void ImportImage(FvsImage_t image, const FvsString_t filename)
{
    FvsFile_t file = FileCreate();
    if (file!=NULL)
    {
        fprintf(stdout, "Opening file %s...\n", filename);
        if (FileOpen(file, filename, FvsFileRead)==FvsOK)
        {
            (void)AutoImport(file, image);
            (void)FileClose (file);
        }
        FileDestroy(file);
    }
}

static FvsError_t OverlayDirection(FvsImage_t image, const FvsFloatField_t field)
{
    FvsError_t nRet = FvsOK;
    FvsInt_t w      = ImageGetWidth (image);
    FvsInt_t h      = ImageGetHeight(image);
    FvsInt_t pitch, dirp;
    FvsFloat_t theta, c, s;
    FvsByte_t* p;
    FvsFloat_t* orientation;
    FvsInt_t x, y, size, i, j, l;

    size = 8; /* length in pixels of a vector */

    /* change luminosity */
    (void)ImageLuminosity(image, 168);

    pitch  = ImageGetPitch (image);
    p      = ImageGetBuffer(image);

    orientation = FloatFieldGetBuffer(field);
    dirp        = FloatFieldGetPitch(field);

    if (p==NULL || orientation==NULL)
        return FvsMemory;

    /* build vectors */
    for (y = size; y < h-size; y+=size-2)
    for (x = size; x < w-size; x+=size-2)
    {
        theta = orientation[x+y*dirp];
        c = cos(theta);
        s = sin(theta);
        /* draw a vector of length size in the direction theta */
        for (l = 0; l < size; l++)
        {
            i = (FvsInt_t)(x + size/2 - l*s);
            j = (FvsInt_t)(y + size/2 + l*c);
            p[i+j*pitch] = 0;
        }
    }
    return nRet;
}

int main(int argc, char *argv[])
{
    FvsImage_t      image = ImageCreate();
    FvsFloatField_t field = FloatFieldCreate();

    /* check if all parameters are there */
    if (argc!=3)
    {
        printf("Usage: fvs input.bmp output.bmp\n");
        FloatFieldDestroy(field);
        ImageDestroy(image);
        return -1;
    }

    if ( (image!=NULL) && (field!=NULL) )
    {
        ImportImage(image, argv[1]);
        /* normalize the image */
        (void)ImageNormalize(image, (FvsByte_t)128, (FvsUint_t)128*128);

        fprintf(stdout, "1/2 Determining the ridge direction\n");
        (void)FingerprintGetDirection(image, field, 7, 2);

        fprintf(stdout, "2/2 Creating a new image with the direction\n");
        (void)OverlayDirection(image, field);
        ExportImage(image, argv[2]);

        fprintf(stdout, "Cleaning up and exiting...\n");
    }
    FloatFieldDestroy(field);
    ImageDestroy(image);
    return 0;
}


