/*########################################################################

  The contents of this file are subject to the Mozilla Public License
  Version 1.0(the "License");   You  may  NOT  use this file except in
  compliance with the License. You may obtain a copy of the License at
                http:// www.mozilla.org/MPL/
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and limitations
  under the License.

  The Initial Developer of the Original Code is Shivang Patel.

  Copyright(C) 2002-2003. All Rights Reserved.

  Authors: Shivang Patel
           Jaap de Haan(jdh)
  
  This file contains functions to read and write BMP files

########################################################################*/

#include <string.h>
#include <stdlib.h>
#include "fmt_bitmap.h"


typedef struct FvsBitmapFileHeader_t
{
    uint16_t    bfType;
    uint32_t    bfSize;
    uint16_t    bfReserved1;
    uint16_t    bfReserved2;
    uint32_t    bfOffBits;
    uint32_t    biSize;
    int32_t     biWidth;
    int32_t     biHeight;
    uint16_t    biPlanes;
    uint16_t    biBitCount;
    uint32_t    biCompression;
    uint32_t    biSizeImage;
    int32_t     biXPelsPerMeter;
    int32_t     biYPelsPerMeter;
    uint32_t    biClrUsed;
    uint32_t    biClrImportant;
} FvsBitmapFileHeader_t;


#define TYPE_BM 0x4d42

/* private prototype definition */
/* import functions */
static FvsError_t Import32BMP          (FvsFile_t in, FvsImage_t out, FvsByte_t* pal);
static FvsError_t Import24BMP          (FvsFile_t in, FvsImage_t out, FvsByte_t* pal);
static FvsError_t Import16BMP          (FvsFile_t in, FvsImage_t out, FvsByte_t* pal);
static FvsError_t Import4BMP           (FvsFile_t in, FvsImage_t out, FvsByte_t* pal);
static FvsError_t Import8BMP           (FvsFile_t in, FvsImage_t out, FvsByte_t* pal);
static FvsError_t Import4RLE           (FvsFile_t in, FvsImage_t out, FvsByte_t* pal);
static FvsError_t Import8RLE           (FvsFile_t in, FvsImage_t out, FvsByte_t* pal);
static FvsError_t BitmapReadPalette    (FvsFile_t in, /*@out@*/ FvsByte_t* pal, FvsBitmapFileHeader_t* bmpheader);
static FvsError_t BitmapReadHeader     (FvsFile_t in, /*@out@*/ FvsBitmapFileHeader_t* bmpheader);

/* export functions */
static FvsError_t Export8BMP           (FvsFile_t out, const FvsImage_t img);
static FvsError_t BitmapWritePalette   (FvsFile_t out);
static FvsError_t BitmapWriteHeader    (FvsFile_t out, /*@in@*/ FvsBitmapFileHeader_t* bmpheader);


FvsError_t BitmapImport(FvsFile_t in, FvsImage_t img)
{
    FvsError_t nRet = FvsOK;
    FvsBitmapFileHeader_t  bmpheader;
    FvsByte_t              pal[256];

    memset(&bmpheader, 0, sizeof(bmpheader));
    if (FileIsOpen(in)==FvsTrue)
    {
        nRet = FileSeekToBegin(in);
        if (nRet==FvsOK)
            nRet = BitmapReadHeader(in, &bmpheader);
        if (nRet==FvsOK)
        {
            if (bmpheader.bfType!=TYPE_BM ||
                bmpheader.biSize!=40)
                nRet = FvsBadFormat;
            else
            {
                nRet = BitmapReadPalette(in, pal, &bmpheader);
                if (nRet==FvsOK)
                {
                    /* jump to the bitmap bits */
                    nRet = FileSeek(in, (FvsUint_t)bmpheader.bfOffBits);
                    if (nRet==FvsOK)
                    {
                        nRet = ImageSetSize(img,
                                  (FvsInt_t)bmpheader.biWidth,
                                  (FvsInt_t)bmpheader.biHeight);
                        if (nRet==FvsOK)
                        {
                        switch(bmpheader.biBitCount)
                        {
                        case 4:
                            if (bmpheader.biCompression==0)
                                nRet = Import4BMP(in, img, pal);
                            else
                                nRet = Import4RLE(in, img, pal);
                            break;
                        case 8:
                            if (bmpheader.biCompression==0)
                                nRet = Import8BMP(in, img, pal);
                            else
                                nRet = Import8RLE(in, img, pal);
                            break;
                        case 16: nRet = Import16BMP(in, img, pal); break;
                        case 24: nRet = Import24BMP(in, img, pal); break;
                        case 32: nRet = Import32BMP(in, img, pal); break;

                        case 1:
                        default:
                            nRet = FvsBadFormat;
                            break;
                        }
                        }
                    }
                }
                else
                    nRet = FvsBadFormat;
            }
        }
        else
            nRet = FvsBadFormat;
    }
    else
        nRet = FvsBadParameter;
    return nRet;
}


FvsError_t BitmapExport(FvsFile_t out, const FvsImage_t img)
{
    uint32_t len;
    FvsError_t nRet = FvsOK;
    FvsBitmapFileHeader_t  bmpheader;

    if (FileIsOpen(out)==FvsTrue)
    {
        nRet = FileSeekToBegin(out);
        if (nRet!=FvsOK)
            return nRet;
        len = sizeof(bmpheader);
        memset(&bmpheader, 0, sizeof(bmpheader));
        bmpheader.bfType        = TYPE_BM;
        bmpheader.biSize        = 40;
        bmpheader.biWidth       = (int32_t)ImageGetWidth(img);
        bmpheader.biHeight      = (int32_t)ImageGetHeight(img);
        bmpheader.biPlanes      = 1;
        bmpheader.biBitCount    = 8;
        bmpheader.biClrImportant= 256;
        bmpheader.biClrUsed     = 256;
        bmpheader.biSizeImage   = (uint32_t)ImageGetSize(img);
        nRet = BitmapWriteHeader(out, &bmpheader);
        if (nRet==FvsOK)
        {
            nRet = BitmapWritePalette(out);
            if (nRet==FvsOK)
            {
                bmpheader.bfOffBits = (uint32_t)FileGetPosition(out);
                nRet = Export8BMP(out, img);
                bmpheader.bfSize = (uint32_t)FileGetPosition(out);
                if (nRet==FvsOK)
                {
                    (void)FileSeekToBegin(out);
                    nRet = BitmapWriteHeader(out, &bmpheader);
                }
            }
        }
        else
            nRet = FvsIoError;
    }
    else
        nRet = FvsBadParameter;
    return nRet;
}


static FvsError_t Import32BMP(FvsFile_t in, FvsImage_t img, FvsByte_t* pal)
{
    FvsError_t nRet = FvsOK;
    FvsByte_t*  pDest = ImageGetBuffer(img);
    FvsInt_t    nMax  = ImageGetSize(img);
    FvsByte_t   r,g,b;
    FvsInt_t    i;
    if (pDest==NULL)
        return FvsMemory;
    for (i = 0; i < nMax; i++)
    {
        r = FileGetByte(in);
        g = FileGetByte(in);
        b = FileGetByte(in);
        (void)FileGetByte(in); /* skipping */
        *pDest++ = pal[((r+(g<<1)+b)>>2)&0xFF];
    }
    return nRet;
}


static FvsError_t Import24BMP(FvsFile_t in, FvsImage_t img, FvsByte_t* pal)
{
    FvsError_t nRet = FvsOK;
    FvsByte_t*  pDest = ImageGetBuffer(img);
    FvsInt_t    nMax  = ImageGetSize(img);
    FvsByte_t   r,g,b;
    FvsInt_t    i;
    if (pDest==NULL)
        return FvsMemory;
    for (i = 0; i < nMax; i++)
    {
        r = FileGetByte(in);
        g = FileGetByte(in);
        b = FileGetByte(in);
        *pDest++ = pal[((r+(g<<1)+b)>>2)&0xFF];
    }
    return nRet;
}


static FvsError_t Import16BMP(FvsFile_t in, FvsImage_t img, FvsByte_t* pal)
{
    FvsError_t nRet = FvsOK;
    FvsByte_t*  pDest = ImageGetBuffer(img);
    FvsInt_t    nMax  = ImageGetSize(img);
    uint16_t    w;
    FvsInt_t    i;
    if (pDest==NULL)
        return FvsMemory;
    for (i = 0; i < nMax; i++)
    {
        w = FileGetWord(in);
        *pDest++ = pal[( ((w&0xF800)>>11)+((w&0x07E0)>>5)+(w&0x001F))&0xFF];
    }
    return nRet;
}


static FvsError_t Import4BMP(FvsFile_t in, FvsImage_t img, FvsByte_t* pal)
{
    FvsError_t nRet = FvsOK;
    FvsByte_t*  pDest = ImageGetBuffer(img);
    FvsInt_t    nMax  = ImageGetSize(img);
    FvsByte_t   b;
    FvsInt_t    i;
    if (pDest==NULL)
        return FvsMemory;
    for (i = 0; i < nMax; i+=2)
    {
        b = FileGetByte(in);
        *pDest++ = pal[(b>>4)&0x0F];
        *pDest++ = pal[b&0xF];
    }
    return nRet;
}


static FvsError_t Import8BMP(FvsFile_t in, FvsImage_t img, FvsByte_t* pal)
{
    FvsError_t nRet = FvsOK;
    FvsByte_t* pDest = ImageGetBuffer(img);
    FvsInt_t   nMax  = ImageGetSize(img);
    FvsByte_t  b;
    FvsInt_t   i;

    if (pDest==NULL)
    {
        nRet = FvsMemory;
    }
    else
    {
        for (i = 0; i < nMax; i++)
        {
            b = FileGetByte(in);
            *pDest++ = pal[b];
        }
    }
    return nRet;
}


static FvsError_t Export8BMP(FvsFile_t out, const FvsImage_t img)
{
    FvsUint_t size = (FvsUint_t)ImageGetSize(img);
    FvsByte_t* buffer = ImageGetBuffer(img);
    if (buffer==NULL)
        return FvsMemory;
    return (FileWrite(out, buffer, size)==size)?FvsOK:FvsIoError;
}


static FvsError_t Import4RLE(FvsFile_t in, FvsImage_t img, FvsByte_t* pal)
{
    FvsError_t nRet = FvsOK;
    FvsByte_t*  pDest = ImageGetBuffer(img);
    FvsInt_t    nMax  = ImageGetSize(img);
    FvsBool_t   bEOL  = FvsFalse;
    FvsBool_t   bEOF  = FvsFalse;
    FvsByte_t   nLen;
    FvsInt_t    h = ImageGetHeight(img);
    FvsInt_t    y = 0;

    if (pDest==NULL) return FvsMemory;

    for (y=0; (y<h) && bEOF==FvsFalse; y++)
    {
        bEOL = FvsFalse;
        while (bEOL==FvsFalse && bEOF==FvsFalse &&
               FileIsAtEOF(in)==FvsFalse && nMax>0)
        { 
            nLen = FileGetByte(in);
            if (nLen==0)
            { 
                nLen = FileGetByte(in);
                switch (nLen)
                {
                case 1: /* End of file escape ( 0x00 0x01 ) */
                case 2: /* treat seldom used delta escape the same way... */
                    bEOF = FvsTrue;
                    bEOL = FvsTrue;
                    break;
                    /* fallthrough */
                case 0: /* End of line escape ( 0x00 0x00 ) */
                    bEOL = FvsTrue;
                    break;
                default:
                    nMax-=nLen;
                    if (nMax>=0)
                    {
                        FvsByte_t i;
                        /* Literal */
                        for (i = 0; i<nLen;)
                        {
                            FvsByte_t b = FileGetByte(in);
                            *pDest++=pal[b>>4];
                            i++;
                            if (i<nLen)
                            {
                                *pDest++=pal[b&0xF];
                                i++;
                            }
                        }
                        /* Word alignment at end of literal packet. */
                        if ((nLen & 1)!=0)
                            (void)FileGetByte(in); /* ignore */
                    }
                    break;
                }
            }
            else
            {
                nMax-=nLen;
                if (nMax>=0)
                {
                    /* RunLength pixels, all with the same value */
                    FvsByte_t val = pal[FileGetByte(in)];
                    memset(pDest, (int)val, nLen);
                    pDest+=nLen;
                }
            }
        }
    }
    return nRet;
}


FvsError_t Import8RLE(FvsFile_t in, FvsImage_t img, FvsByte_t* pal)
{
    FvsError_t nRet = FvsOK;
    FvsByte_t*  pDest = ImageGetBuffer(img);
    FvsBool_t   bEOL = FvsFalse;
    FvsBool_t   bEOF = FvsFalse;
    FvsByte_t   nLen, i;
    FvsInt_t    w = ImageGetWidth(img);
    FvsInt_t    h = ImageGetHeight(img);
    FvsInt_t    nMax = h*w;
    FvsByte_t   val = 0;
    FvsInt_t    y = 0;

    if (pDest==NULL) return FvsMemory;

    for (y=0; (y<h) && bEOF==FvsFalse; y++)
    {
        bEOL = FvsFalse;
	    /* start of line */
	    while (bEOL==FvsFalse && bEOF==FvsFalse &&
               FileIsAtEOF(in)==FvsFalse && nMax>0)
        { 
            nLen = FileGetByte(in);
            if (nLen==0)
            { 
                nLen = FileGetByte(in);
                switch (nLen)
                {
                case 1: /* End of file escape ( 0x00 0x01 ) */
                case 2: /* treat seldom used delta escape the same way...*/
		            bEOF = FvsTrue;
                    bEOL = FvsTrue;
                    break;
                    
		        /* fallthrough */		    
                case 0: /* End of line escape ( 0x00 0x00 ) */
                    bEOL = FvsTrue;
                    break;
                default:
                    nMax-=nLen;
                    if (nMax>=0)
                    {
                        
                        /* Literal */
                        for (i = 0; i<nLen; i++)
                            *pDest++=pal[FileGetByte(in)];
                        /* Word alignment at end of literal packet. */
                        if ((nLen & 1)!=0)
                            (void)FileGetByte(in); /* skip */
                    }
                    break;
                }
            }
            else
            {
                nMax-=nLen;
                if (nMax>=0)
                {
                    /* RunLength pixels, all with the same value */
                    val = pal[FileGetByte(in)];
        	        memset(pDest, (int)val, nLen);		    
                    pDest+=nLen;
                }
            }
        }
    }
    return nRet;
}


/*
** Write the palette into the file
*/
static FvsError_t BitmapWritePalette(FvsFile_t out)
{
    FvsUint_t nBytes;
    FvsUint_t nColors = 256;
    FvsUint_t i;
    FvsUint_t _pal[256];
    memset(_pal, 0, 256);
    for (i=0; i<nColors; i++)
        _pal[i] = (( (i<<8) + i ) << 8) + i;

    nBytes = FileWrite(out, _pal, nColors*4);
    return (nBytes==nColors*4)?FvsOK:FvsIoError;
}


/*
** Adapter to read in the header without complicated packing options
** and relying on memory alignment. This was already present in the
** first version of the bitmap import function from Shivang Patel.
*/
static FvsError_t BitmapReadHeader(FvsFile_t in, FvsBitmapFileHeader_t* bmpheader)
{
    if (FileRead(in, &bmpheader->bfType, 2)!=2)
        return FvsIoError;
    if (FileRead(in, &bmpheader->bfSize, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->bfReserved1, 2)!=2)
        return FvsIoError;
    if (FileRead(in, &bmpheader->bfReserved2, 2)!=2)
        return FvsIoError;
    if (FileRead(in, &bmpheader->bfOffBits, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biSize, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biWidth, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biHeight, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biPlanes, 2)!=2)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biBitCount, 2)!=2)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biCompression, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biSizeImage, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biXPelsPerMeter, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biYPelsPerMeter, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biClrUsed, 4)!=4)
        return FvsIoError;
    if (FileRead(in, &bmpheader->biClrImportant, 4)!=4)
        return FvsIoError;
    return FvsOK;
}


/*
** Adpater function to write out the header.
*/
static FvsError_t BitmapWriteHeader(FvsFile_t out, FvsBitmapFileHeader_t* bmpheader)
{
    if (FileWrite(out, &bmpheader->bfType, 2)!=2)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->bfSize, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->bfReserved1, 2)!=2)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->bfReserved2, 2)!=2)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->bfOffBits, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biSize, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biWidth, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biHeight, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biPlanes, 2)!=2)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biBitCount, 2)!=2)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biCompression, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biSizeImage, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biXPelsPerMeter, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biYPelsPerMeter, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biClrUsed, 4)!=4)
        return FvsIoError;
    if (FileWrite(out, &bmpheader->biClrImportant, 4)!=4)
        return FvsIoError;
    return FvsOK;
}


/*
** Read in the palette, in order to do this correctly for all formats, we
** need some information from the bitmap header.
*/
static FvsError_t BitmapReadPalette(FvsFile_t in, FvsByte_t* pal, FvsBitmapFileHeader_t* bmpheader)
{
    FvsError_t nRet = FvsOK;

    FvsUint_t nColors = 0;
    FvsUint_t i;

    /* default palette */
    for (i = 0; i < 256; i++)
        pal[i] = (FvsByte_t)i;

    switch (bmpheader->biBitCount)
    {
    case 1:     case 4:     case 8:
        nColors = (bmpheader->biClrUsed!=0)?
            (FvsUint_t)bmpheader->biClrUsed:
            (FvsUint_t)(1<<bmpheader->biBitCount);
        break;
    case 16:    case 24:    case 32:
        /* OK */
        break;
    default:
        nRet = FvsBadFormat;
    }

    if (nColors!=0)
    {
        FvsUint_t _pal[256];
        FvsUint_t nBytes = FileRead(in, _pal, nColors*4);
        if (nBytes==nColors*4)
        {
            /* Correct the byte ordering & copy the data. */
            for (i=0; i<nColors; i++)
            {
                FvsUint_t r = (_pal[i])&0xFF;
                FvsUint_t g = (_pal[i]>>8)&0xFF;
                FvsUint_t b = (_pal[i]>>16)&0xFF;
                pal[i] = (FvsByte_t)(((r+(g<<1)+b)>>2)&0xFF);
            }
        }
        else
            nRet = FvsBadFormat;
    }
    return nRet;
}


